// (C) 2003-2015 Doctor Web. All rights reserved.

#include "drweb_netcheck_api.h"

#include <stdexcept>

using namespace dw;
using namespace nc;

std::shared_ptr<netcheck_api_t> netcheck_api_t::create(boost::asio::io_service& io)
{
    return std::shared_ptr<netcheck_api_t>(new netcheck_api_t(io));
}

netcheck_api_t::netcheck_api_t(boost::asio::io_service& io) : ipc_t(io), serial_(0), window_size_(0)
{
}

void netcheck_api_t::start(const std::string& path)
{
    async_connect(path);
    send_hello();
}

void netcheck_api_t::stop()
{
    for(const auto& job : jobs_) send_reset(job.first);
    jobs_.clear();
    close_ipc();
}

uint32_t netcheck_api_t::init_job(delegate_t* delegate)
{
    const uint32_t job = serial_++;
    jobs_.emplace(job, delegate);
    ClientMessage msg;
    msg.set_tag(ClientMessage::INIT);
    msg.set_job_id(job);
    async_send_message(msg);
    return job;
}

void netcheck_api_t::send_data(uint32_t job, const void* data, size_t size)
{
    ClientMessage msg;
    msg.set_tag(ClientMessage::DATA_CHUNK);
    msg.set_job_id(job);
    msg.set_data(data, size);
    async_send_message(msg);
    window_size_ -= size;
    assert(window_size_ >= 0);
}

void netcheck_api_t::send_scan_request(uint32_t job, const ScanOptions& options)
{
    ClientMessage msg;
    msg.set_tag(ClientMessage::SCAN_REQUEST);
    msg.set_job_id(job);
    msg.mutable_scan_options()->CopyFrom(options);
    async_send_message(msg);
}

void netcheck_api_t::send_get_data_request(uint32_t job)
{
    ClientMessage msg;
    msg.set_tag(ClientMessage::GET_DATA_REQUEST);
    msg.set_job_id(job);
    async_send_message(msg);
}

void netcheck_api_t::send_reset(uint32_t job)
{
    ClientMessage msg;
    msg.set_tag(ClientMessage::RESET);
    msg.set_job_id(job);
    async_send_message(msg);
    jobs_.erase(job);
}

void netcheck_api_t::message_received(google::protobuf::io::ZeroCopyInputStream* is)
{
    ServerMessage msg;
    if(!msg.ParseFromZeroCopyStream(is) || !msg.IsInitialized())
        throw std::domain_error("bad message format");

    assert(msg.has_tag());
    const uint32_t job = msg.job_id();
    switch(msg.tag()) {
        case ServerMessage::SCAN_RESPONSE:
            assert(msg.has_scan_report());
            assert(msg.has_job_id());
            handle_scan_response(job, msg.scan_report());
            break;
        case ServerMessage::GET_DATA_RESPONSE:
            assert(msg.has_data_size());
            assert(msg.has_job_id());
            handle_get_data_response(job, msg.data_size());
            break;
        case ServerMessage::DATA_CHUNK:
            assert(msg.has_data());
            assert(msg.has_job_id());
            handle_data_chunk(job, msg.data());
            break;
        case ServerMessage::ALARM:
            assert(msg.has_error());
            assert(msg.has_job_id());
            handle_alarm(job, msg.error().code());
            break;
        case ServerMessage::ADJUST_WINDOW: {
            assert(msg.has_window_change());
            bool replenished = window_size_ <= 0;
            window_size_ += msg.window_change();
            replenished = replenished && window_size_ > 0;
            if(replenished) {
                for(auto& delegate : jobs_) {
                    delegate.second->on_window_replenished();
                    assert(window_size_ >= 0);
                    if(window_size_ == 0) break;
                }
            }
            break;
        }
        default:
            throw std::domain_error("unexpected message");
    }
}

void netcheck_api_t::ipc_error_occurred(const boost::system::error_code& ec)
{
    throw boost::system::system_error(ec, "netcheck_ipc");
}

std::shared_ptr<ipc_t> netcheck_api_t::this_as_ipc()
{
    return shared_from_this();
}

void netcheck_api_t::send_hello()
{
    ClientMessage msg;
    msg.set_tag(ClientMessage::HELLO);
    msg.set_origin(APP_GENERIC);
    msg.set_origin_pid(getpid());
    msg.set_flow_control(true);
    async_send_message(msg);
}

void netcheck_api_t::handle_scan_response(uint32_t job, const ScanReport& report)
{
    const auto found = jobs_.find(job);
    if(found != jobs_.end()) found->second->on_scan_report(job, report);
}

void netcheck_api_t::handle_get_data_response(uint32_t job, off_t size)
{
    const auto found = jobs_.find(job);
    if(found != jobs_.end()) found->second->on_cured_data_size(job, size);
}

void netcheck_api_t::handle_data_chunk(uint32_t job, const std::string& data)
{
    const auto found = jobs_.find(job);
    if(found != jobs_.end()) found->second->on_cured_data(job, data.c_str(), data.size());
}

void netcheck_api_t::handle_alarm(uint32_t job, ErrorCode ec)
{
    const auto found = jobs_.find(job);
    if(found != jobs_.end()) found->second->on_alarm(job, ec);
}
