#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""
    An example of drweb-netcheck API usage in synchronous mode with flow control support.
"""

import os
import socket
import struct
import sys

import drweb_netcheck_api_pb2 as netcheck

SIZE_STRUCT = '<I'
files = [ '/bin/sh', '/bin/ls', '/bin/rm' ]

def get_ncheck_socket():
    if len(sys.argv) != 2:
        print('Usage: {} <socket path>'.format(sys.argv[0]))
        exit(1)
    return sys.argv[1]

class netcheck_ipc(object):
    def __init__(self, socket_path):
        self.serial = 0
        self.window_size = 0
        self.sock = socket.socket(socket.AF_UNIX, socket.SOCK_STREAM)
        self.sock.connect(socket_path)
        self.send_hello()

    def send_hello(self):
        message = netcheck.ClientMessage()
        message.tag = message.HELLO
        message.origin = netcheck.APP_GENERIC
        message.origin_pid = os.getpid()
        message.flow_control = True
        self.send_message(message)
        self.replenish_window()

    def replenish_window(self):
        while self.window_size <= 0:
            msg = self.recv_message()
            if msg.tag == netcheck.ServerMessage().ALARM:
                print('Received alarm message for job {}'.format(job_id))
            elif msg.tag == netcheck.ServerMessage().ADJUST_WINDOW:
                assert msg.HasField('window_change')
                self.window_size += msg.window_change
            else:
                raise Exception('Unexpected message from NetCheck')

    def scan(self, path):
        job_id = self.send_init()
        self.send_file_by_chunks(job_id, path)
        self.send_scan(job_id)
        self.process_response()

    def send_init(self):
        job_id = self.serial
        self.serial += 1
        message = netcheck.ClientMessage()
        message.tag = message.INIT
        message.job_id = job_id
        self.send_message(message)
        return job_id

    def send_file_by_chunks(self, job_id, path):
        fin = open(path, 'rb')
        while True:
            self.replenish_window()
            assert self.window_size > 0

            chunk = fin.read(min(self.window_size, 4096))
            if not chunk:
                break
            self.send_data_chunk(job_id, chunk)
        fin.close()

    def send_data_chunk(self, job_id, chunk):
        message = netcheck.ClientMessage()
        message.tag = message.DATA_CHUNK
        message.job_id = job_id
        message.data = chunk
        self.send_message(message)
        self.window_size -= len(chunk)
        assert self.window_size >= 0

    def send_scan(self, job_id):
        message = netcheck.ClientMessage()
        message.tag = message.SCAN_REQUEST
        message.job_id = job_id
        message.scan_options.CopyFrom(netcheck.ScanOptions())
        message.scan_options.cure = True
        self.send_message(message)

    def send_get_data_request(self, job_id):
        message = netcheck.ClientMessage()
        message.tag = message.GET_DATA_REQUEST
        message.job_id = job_id
        self.send_message(message)

    def send_reset(self, job_id):
        message = netcheck.ClientMessage()
        message.tag = message.RESET
        message.job_id = job_id
        self.send_message(message)

    def process_response(self):
        while True:
            msg = self.recv_message()
            if msg.tag == netcheck.ServerMessage().SCAN_RESPONSE:
                assert msg.HasField('scan_report')
                assert msg.HasField('job_id')
                report = msg.scan_report
                print(report)
                if report.cured and not report.cured_by_deletion:
                    self.recv_cured_data(msg.job_id)
                self.send_reset(msg.job_id)
                return
            elif msg.tag == netcheck.ServerMessage().ALARM:
                assert msg.HasField('job_id')
                raise Exception('Received alarm message for job {}'.format(msg.job_id))
                return
            elif msg.tag == netcheck.ServerMessage().ADJUST_WINDOW:
                assert msg.HasField('window_change')
                self.window_size += msg.window_change
            else:
                raise Exception('Unexpected message from NetCheck')

    def recv_cured_data(self, job_id):
        self.send_get_data_request(job_id)
        size = self.recv_get_data_response(job_id)
        cured_data = ""
        while len(cured_data) < size:
            buffer = self.recv_data_chunk(job_id)
            cured_data += buffer
            print("Cured data received job_id: {0}, size: {1}, remains: {2}".\
                format(job_id, len(buffer), size - len(cured_data)))

    def recv_get_data_response(self, job_id):
        response = self.recv_message()
        assert response.tag == netcheck.ServerMessage().GET_DATA_RESPONSE
        assert response.HasField('data_size')
        assert response.job_id == job_id
        return response.data_size

    def recv_data_chunk(self, job_id):
        data_chunk = self.recv_message()
        assert data_chunk.tag == netcheck.ServerMessage().DATA_CHUNK
        assert data_chunk.HasField('data')
        assert data_chunk.job_id == job_id
        return data_chunk.data

    def recv_message(self):
        size = struct.unpack(SIZE_STRUCT,
                             self.sock.recv(struct.calcsize(SIZE_STRUCT)))
        size = size[0]
        buffer = b''
        while len(buffer) < size:
            buffer += self.sock.recv(size - len(buffer))

        message = netcheck.ServerMessage()
        message.ParseFromString(buffer)
        return message

    def send_message(self, message):
        assert message.IsInitialized()
        buffer = struct.pack(SIZE_STRUCT, message.ByteSize())
        buffer += message.SerializeToString()
        while len(buffer):
            send = self.sock.send(buffer)
            buffer = buffer[send:]

ipc = netcheck_ipc(get_ncheck_socket())
for f in files:
    ipc.scan(f)
