// An example of drweb-netcheck API usage in synchronous mode without flow control support.
// (C) 2003-2015 Doctor Web. All rights reserved.

#include <errno.h>
#include <fcntl.h>
#include <stdlib.h>
#include <sys/socket.h>
#include <sys/stat.h>
#include <sys/types.h>
#include <sys/un.h>
#include <unistd.h>

#include <iostream>
#include <sstream>

#include <google/protobuf/io/coded_stream.h>
#include <google/protobuf/io/zero_copy_stream_impl_lite.h>

#include "drweb_netcheck_api.pb.h"

using namespace dw;
using namespace nc;

void write_exact(int fd, const void* const buffer, size_t size)
{
    ssize_t total = 0;
    do {
        ssize_t part = write(fd, static_cast<const char*>(buffer) + total, size - total);
        if(part > 0) {
            total += part;
        } else if(part == 0 && size > 0) {
            throw std::runtime_error("write() failed");
        } else if(part == -1) {
            if(errno == EINTR) continue;
            throw std::runtime_error("write() failed");
        }
    } while(total > 0 && (size_t)total < size);
}

void read_exact(int fd, void* buffer, size_t size)
{
    size_t total = 0;
    do {
        ssize_t part = read(fd, static_cast<char*>(buffer) + total, size - total);
        if(part > 0) {
            total += part;
        } else if(part == 0 && size > 0) {
            throw std::runtime_error("read() failed");
        } else if(part == -1) {
            if(errno == EINTR) continue;
            throw std::runtime_error("read() failed");
        }
    } while(total < size);
}

void send_message(int fd, const google::protobuf::Message& msg)
{
    static const size_t header_size = sizeof(uint32_t);
    const size_t message_size = msg.ByteSize();

    using namespace google::protobuf::io;
    std::vector<uint8_t> buffer;
    buffer.resize(header_size + message_size);
    CodedOutputStream::WriteLittleEndian32ToArray(message_size, &buffer[0]);
    msg.SerializeWithCachedSizesToArray(&buffer[header_size]);

    write_exact(fd, &buffer[0], buffer.size());
}

void recv_message(int fd, google::protobuf::Message* msg)
{
    static const size_t header_size = sizeof(uint32_t);
    std::vector<uint8_t> buffer(header_size);
    read_exact(fd, &buffer[0], buffer.size());

    using namespace google::protobuf::io;
    uint32_t message_size;
    CodedInputStream::ReadLittleEndian32FromArray(&buffer[0], &message_size);
    buffer.resize(message_size);
    read_exact(fd, &buffer[0], buffer.size());
    ArrayInputStream stream(&buffer[0], buffer.size());
    if(!msg->ParseFromZeroCopyStream(&stream)) throw std::runtime_error("bad message format");
}

int connect_to_socket(const char* path)
{
    int fd = socket(AF_LOCAL, SOCK_STREAM, 0);
    if(fd == -1) throw std::runtime_error("socket() failed");

    sockaddr_un addr;
    addr.sun_family = AF_LOCAL;
    strncpy(addr.sun_path, path, sizeof(addr.sun_path));
    if(connect(fd, reinterpret_cast<sockaddr*>(&addr), sizeof(addr)) == -1) {
        close(fd);
        throw std::runtime_error("connect() failed");
    }

    return fd;
}

void send_hello(int fd)
{
    ClientMessage msg;
    msg.set_tag(ClientMessage::HELLO);
    msg.set_origin(APP_GENERIC);
    msg.set_origin_pid(getpid());
    send_message(fd, msg);
}

void send_init(int fd, uint32_t job_id)
{
    ClientMessage msg;
    msg.set_tag(ClientMessage::INIT);
    msg.set_job_id(job_id);
    send_message(fd, msg);
}

void send_data_chunk(int fd, uint32_t job_id, const void* data, size_t size)
{
    ClientMessage msg;
    msg.set_tag(ClientMessage::DATA_CHUNK);
    msg.set_job_id(job_id);
    msg.set_data(data, size);
    send_message(fd, msg);
}

void send_file_by_chunks(int fd, uint32_t job_id, const char* file)
{
    int file_fd = open(file, O_RDONLY);
    if(file_fd == -1) throw std::runtime_error("open() failed");

    for(;;) {
        char buffer[16384];
        ssize_t part = read(file_fd, buffer, sizeof(buffer));
        if(part == 0) break;
        if(part == -1) throw std::runtime_error("read() failed");
        send_data_chunk(fd, job_id, buffer, part);
    }

    close(file_fd);
}

void send_reset(int fd, uint32_t job_id)
{
    ClientMessage msg;
    msg.set_tag(ClientMessage::RESET);
    msg.set_job_id(job_id);
    send_message(fd, msg);
}

void send_scan_request(int fd, uint32_t job_id, const ScanOptions& options)
{
    ClientMessage msg;
    msg.set_tag(ClientMessage::SCAN_REQUEST);
    msg.set_job_id(job_id);
    msg.mutable_scan_options()->CopyFrom(options);
    send_message(fd, msg);
}

void send_get_data_request(int fd, uint32_t job_id)
{
    ClientMessage msg;
    msg.set_tag(ClientMessage::GET_DATA_REQUEST);
    msg.set_job_id(job_id);
    send_message(fd, msg);
}

void process_answer(int fd, uint32_t job_id)
{
    off_t remains = 0;
    for(;;) {
        ServerMessage msg;
        recv_message(fd, &msg);

        switch(msg.tag()) {
            case ServerMessage::SCAN_RESPONSE:
                std::cout << msg.scan_report().DebugString() << std::endl;
                if(!msg.scan_report().cured() || msg.scan_report().cured_by_deletion()) return;
                send_get_data_request(fd, job_id);
                break;
            case ServerMessage::GET_DATA_RESPONSE:
                remains = msg.data_size();
                std::cout << "Cured data size: " << remains << std::endl;
                break;
            case ServerMessage::DATA_CHUNK:
                assert(static_cast<off_t>(msg.data().size()) <= remains);
                remains -= msg.data().size();
                std::cout << "Received cured data chunk: " << msg.data().size()
                          << ", remains: " << remains << std::endl;
                if(!remains) return;
                break;
            case ServerMessage::ALARM:
                throw std::runtime_error("unexpected job alarm");
            default:
                throw std::runtime_error("unexpected message from NetCheck");
        }
    }
}

void scan(const std::string& file, const std::string& socket_path)
{
    const uint32_t job_id = 0;
    int fd = connect_to_socket(socket_path.c_str());
    send_hello(fd);
    send_init(fd, job_id);
    send_file_by_chunks(fd, job_id, file.c_str());

    ScanOptions options;
    options.set_cure(true);
    send_scan_request(fd, job_id, options);

    process_answer(fd, job_id);
    send_reset(fd, job_id);
    close(fd);
}

int main(int argc, char* argv[])
{
    try {
        if(argc != 2) {
            std::ostringstream os;
            os << "Usage: " << argv[0] << " <socket path>";
            throw std::runtime_error(os.str());
        }
        std::string socket_path = argv[1];
        scan("/bin/sh", socket_path);
        scan("/bin/ls", socket_path);
        scan("/bin/rm", socket_path);
    } catch(const std::exception& ex) {
        std::cerr << ex.what() << std::endl;
        return EXIT_FAILURE;
    }
}
